<?php

namespace Drupal\semantic_search\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\node\Entity\Node;
use Algenza\Cosinesimilarity\Cosine;
use Drupal\semantic_search\Services\EmbeddingStorageService;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Pgvector\Vector;

/**
 * SearchForm form class.
 */
class SearchForm extends FormBase {

  /**
   * The Embedding Storage client.
   *
   * @var \Drupal\semantic_search\Services\EmbeddingStorageService
   */
  protected $storage_client;

  /**
   * The Embedding Vectorization client.
   *
   * @var \Drupal\semantic_search\Services\EmbeddingVectorizationService
   */
  protected $vectorizing_client;

  /**
   * The Pinecone HTTP client.
   *
   * @var \Drupal\semantic_search\Http\PineconeClient
   */
  protected $pinecone;

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'semantic_search_form';
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);
    $instance->storage_client = $container->get('semantic_search.embedding_storage');
    $instance->vectorizing_client = $container->get('semantic_search.embedding_vectorization');
    $instance->pinecone = $container->get('semantic_search.pinecone_client');
    $instance->entityTypeManager = $container->get('entity_type.manager');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $form['semantic_search'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Search'),
      '#required' => TRUE,
    ];

    $form['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Submit'),
    ];

    if (isset($form_state->getUserInput()['semantic_search']) && $query = $form_state->getUserInput()['semantic_search']) {

      $similarity_min = \Drupal::config('semantic_search.settings')->get('similarity_min');

      $embedding_service = \Drupal::service('semantic_search.embedding_vectorization');
      $search_embedding = $embedding_service->getEmbedding($query);

      $result_ids = [];

      switch (\Drupal::config('semantic_search.settings')->get('embedding_storage')) {
        case EmbeddingStorageService::LOCAL_STORAGE_KEY:


          $storage_service = \Drupal::service('semantic_search.embedding_storage');
          $embeddings = $storage_service->getSavedEmbeddings();

          foreach ($embeddings as $key => $embedding) {
            $cosine_similarity = Cosine::similarity($search_embedding, $embedding);
            if ($cosine_similarity > floatval($similarity_min)) {
              $similarities[$key] = $cosine_similarity;
            }
          }

          if ($similarities) {

            arsort($similarities);

            foreach ($similarities as $key => $similarity) {
              $result_ids[$key] = $key;
            }

          }
          break;
        case EmbeddingStorageService::PINECONE_STORAGE_KEY:

          $result = $this->vectorizing_client->getEmbedding($query);

          $pinecone_query = $this->pinecone->query(
            $result,
            8,
            TRUE,
            FALSE,
            [
            ],
            EmbeddingStorageService::PINECONE_NAMESPACE,
          );

          $result = json_decode($pinecone_query->getBody()->getContents());
          $tracked = [];

          foreach ($result->matches as $match) {
            if (isset($tracked[$match->metadata->entity_type]) && in_array($match->metadata->entity_id, $tracked[$match->metadata->entity_type])) {
              continue;
            }

            if ($match->score < $similarity_min) {
              continue;
            }

            $result_ids[] = $match->metadata->entity_id;

            $tracked[$match->metadata->entity_type][] = $match->metadata->entity_id;
          }

          if (empty($tracked)) {
            $output = '<p>No results were found, or results were excluded because they did not meet the relevancy score threshold.</p>';
          }

          $form['semantic_search_result'][] = ['#markup' => '<div>' . $output . '</div>'];
          break;
        case EmbeddingStorageService::POSTGRES_STORAGE_KEY:

          $output = "";
          $embedding = new Vector($search_embedding);

          $table_name = EmbeddingStorageService::getPostgresTableName();
          $result = EmbeddingStorageService::postgresRequest('SELECT * FROM ' . $table_name . ' ORDER BY embedding <-> $1 LIMIT 5', [$embedding]);

          if ($result) {
            while ($row = pg_fetch_array($result)) {
              $result_ids[] = $row['content'];
            }
          }
          break;
      }

      if (count($result_ids)) {

        $nodes = Node::loadMultiple($result_ids);
        foreach ($nodes as $node) {

          $current_lang = \Drupal::languageManager()->getCurrentLanguage()->getId();
          if ($node->hasTranslation($current_lang)) {
            $node = $node->getTranslation($current_lang);
          }

          // @todo display teaser.
          $form['semantic_search_result'][] = ['#markup' => '<div><a href="' . $node->toUrl()->toString() . '">' . $node->getTitle() . '</a></div>'];
        }
      }
      else {
        $form['semantic_search_result'][] = ['#markup' => '<div>' . $this->t('No match found') . '</div>'];

      }

    }
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {

    $form_state->setRebuild(TRUE);

  }

}
