<?php

namespace Drupal\semantic_search\Services;

use Drupal\Core\File\FileSystemInterface;
use Drupal\semantic_search\Http\PineconeClient;
use Drupal\Core\Database\Connection;
use Pgvector\Vector;

/**
 * Service to handle and store embeddings.
 *
 * @package Drupal\semantic_search\Services
 */
class EmbeddingStorageService {

  /**
   * The embeddings private directory.
   *
   * @var string
   */
  const EMBEDDING_PRIVATE_PATH = 'private://semantic_search/';

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The database connection service.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The Pinecone client.
   *
   * @var \Drupal\semantic_search\Http\PineconeClient
   */
  protected $pinecone;

  /**
   * The Pinecone namespace.
   */
  const PINECONE_NAMESPACE = 'node:rendered_entity';

  /**
   * The Pinecone storage key.
   */
  const PINECONE_STORAGE_KEY = 'pinecone';

  /**
   * The local storage key.
   */
  const LOCAL_STORAGE_KEY = 'local';

  /**
   * The postgres storage key.
   */
  const POSTGRES_STORAGE_KEY = 'postgres';

  /**
   * Constructs for EmbeddingStorageService.
   *
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   */
  public function __construct(FileSystemInterface $file_system, PineconeClient $pinecone_client, Connection $connection) {
    $this->fileSystem = $file_system;
    $this->pinecone = $pinecone_client;
    $this->database = $connection;
  }

  /**
   * Save embeddings to file.
   */
  public function saveEmbeddings($embeddings) {
    $path = $this::EMBEDDING_PRIVATE_PATH;
    $this->fileSystem->prepareDirectory($path, FileSystemInterface::CREATE_DIRECTORY);

    $filepath = $this::getLocalFilepath();
    $this->fileSystem->saveData(json_encode($embeddings, TRUE), $filepath,TRUE);
  }

  public function saveEmbedding($node) {

    $builder = \Drupal::entityTypeManager()->getViewBuilder('node');
    $build = $builder->view($node);
    $node_content = \Drupal::service('renderer')->render($build);

    $embedding_service = \Drupal::service('semantic_search.embedding_vectorization');
    $embedding = $embedding_service->getEmbedding($node_content->jsonSerialize());

    if (!$embedding) {
      return FALSE;
    }

    switch (\Drupal::config('semantic_search.settings')->get('embedding_storage')) {
      case $this::LOCAL_STORAGE_KEY:
        $saved_embeddings = $this->getSavedEmbeddings();
        $saved_embeddings[$node->id()] = $embedding;
        $this->saveEmbeddings($saved_embeddings);
        break;
      case $this::POSTGRES_STORAGE_KEY:
        $vector = new Vector($embedding);
        $table = $this::getPostgresTableName();
        $this::postgresRequest('INSERT INTO ' . $table . ' (embedding, content) VALUES ($1, $2)', [$vector, $node->id()]);
        break;
      case $this::PINECONE_STORAGE_KEY:

        // @todo The entity should be inserted as one string and not several entries
        try {

          $id = 'entity:' . $node->id() . ':' . $node->getEntityTypeId() . ':' . $node->bundle();

          $vectors = [
            'id' => $id,
            'values' => $embedding,
            'metadata' => [
              'entity_id' => $node->id(),
              'entity_type' => $node->getEntityTypeId(),
              'bundle' => $node->bundle()
            ]
          ];

          $this->pinecone->upsert($vectors, $this::PINECONE_NAMESPACE);

          $this->database->merge('semantic_search_embeddings')
            ->keys(
              [
                'entity_id' => $node->id(),
                'entity_type' => $node->getEntityTypeId(),
                'bundle' => $node->bundle(),
              ]
            )
            ->fields(
              [
                'embedding' => json_encode(['data' => $embeddings]) ?? [],
              ]
            )
            ->execute();

          sleep(1);
          usleep(200000);

        } catch (\Exception $e) {
          $this->logger->error(
            'An exception occurred while trying to generate embeddings for a :entity_type with the ID of :entity_id on the :field_name field, with a delta of :field_delta. The bundle of this entity is :bundle. The error was :error',
            [
              ':entity_type' => $entity->getEntityTypeId(),
              ':entity_id' => $entity->id(),
              ':field_name' => $field->getName(),
              ':field_delta' => $delta,
              ':bundle' => $entity->bundle(),
              ':error' => $e->getMessage(),
            ]
          );
        }
        break;
    }

    return TRUE;
  }

  public function removeEmbedding($node) {
    $saved_embeddings = $this->getSavedEmbeddings();

    if (isset($saved_embeddings[$node->id()])) {
      unset($saved_embeddings[$node->id()]);
      $this->saveEmbeddings($saved_embeddings);
    }
  }
  /**
   * Get saved embeddings from file.
   *
   * @return array
   *   Saved embeddings.
   */
  public function getSavedEmbeddings() {

    $saved_embeddings = [];
    $filepath = $this::getLocalFilepath();
    if (file_exists($filepath)) {
      $embedding_data = file_get_contents($filepath);
      $saved_embeddings = json_decode($embedding_data, TRUE);
    }
    return $saved_embeddings;
  }

  public static function getPostgresConnection() {
    $postgres_config = \Drupal::config('semantic_search.postgres_client');
    $db = pg_connect("host=". $postgres_config->get('host') . " dbname=". $postgres_config->get('db_name') . " user=". $postgres_config->get('user') . " password=". $postgres_config->get('password'));
    return $db;
  }

  public static function postgresRequest($query, $args) {
    $db = EmbeddingStorageService::getPostgresConnection();
    $result = pg_query_params($db, $query, $args);
    // pg_free_result($result);
    pg_close($db);
    return $result;
  }

  public static function getPostgresTableName() {
    $vectorization_service = EmbeddingVectorizationService::getVectorizationService();
    $vector_size = EmbeddingVectorizationService::getVectorSize();
    $table_prefix = \Drupal::config('semantic_search.postgres_client')->get('table_prefix');
    return $table_prefix . '_' . $vectorization_service . '_' . $vector_size;
  }

  public static function getLocalFilename() {
    $vectorization_service = EmbeddingVectorizationService::getVectorizationService();
    $vector_size = EmbeddingVectorizationService::getVectorSize();
    return "semantic_search_vectors_" . $vectorization_service . '_' . $vector_size . ".json";
  }

  public static function getLocalFilepath() {
    return EmbeddingStorageService::EMBEDDING_PRIVATE_PATH . EmbeddingStorageService::getLocalFilename();
  }

  public function EraseLocalFile() {
    $filepath = $this::getLocalFilepath();
    $f = @fopen($filepath, "r+");
    if ($f !== false) {
      ftruncate($f, 0);
      fclose($f);
    }
  }

}
